/*
 * Copyright (c) 2007, 2008 Diomidis Spinellis.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Publications using results from this work or its derivatives must
 *    acknowledge the work's origin by citing the corresponding August 2008
 *    Communications of the ACM paper.
 *
 * THIS SOFTWARE IS PROVIDED BY AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: measure.cpp,v 1.38 2008/06/26 06:30:12 dds Exp $
 *
 * Go through a complete XML dump of Wikipedia, measuring the evolution
 * of various attributes.
 */

#include <iostream>
#include <ios>
#include <iomanip>
#include <map>
#include <set>
#include <sstream>
#include <vector>
#include <string>
#include <iterator>
#include <algorithm>

using namespace std;

#include "rangemap.h"

// Processing conditions and results

// All entries (articles)
class EntryDetails {
public:
	RangeMap defined;		// Periods where this entry is defined
	RangeMap stub;			// Periods where this entry is defined as a stub
	vector <int> nrefs;		// Number of references to this entry per period
	string definer;			// First definer (if not an ip address)
	string referrer;		// First referrer (if not an ip address)
	string firstRef;		// Time of first reference
	time_t firstDef;		// Time of first definition
	int numReferences;		// Number of unique contained references to other articles
	int numContributors;		// Number of distinct contributors
	int numRevisions;		// Number of revisions
	int numReverts;			// Number of reverted changes
	EntryDetails(const RangeMap &d, const RangeMap &s) : defined(d), stub(s), nrefs(RangeMap::NMONTH, 0), firstDef(-1),
		numReferences(0), numContributors(0), numRevisions(0), numReverts(0) {}
};

typedef map<string, EntryDetails> EntryMap;
static EntryMap entries;

typedef map<string, RangeMap> ActivityMap;

// All contributors
static ActivityMap contributors;

// Total number of revisions
static int num_revisions;

// Number of pages with restrictions
static int num_restrictions;

// Pages currently valid that are redirected
static int num_redirections;

// Pages with at least one revert comment
static int num_reverted;

// Total number of reverts across all pages
static int num_reverts;

// Total time between an entry and a revert
static long long revert_time;

// Revisions from IP addresses
static int num_ip;

// Number of pages
static int num_pages;

// Maximum number of contributors (and the page)
static int max_contributors;
static string max_contributors_title;

// Maximum number of changes (and the page)
static int max_revisions;
static string max_revisions_title;

// Maximum number of references (and the page)
static int max_references;
static string max_references_title;

// Maximum number of reverts (and the page)
static int max_reverts;
static string max_reverts_title;

// Pages that have moved
class Move {
public:
	string sOld;		// Old name
	string sNew;		// New name
	Move(const string &o, const string &n) : sOld(o), sNew(n) {}
};

typedef vector<Move> MoveContainer;
static MoveContainer moves;

// Temporary variables

static int currently_redirected;
static bool currently_stub;
static string current_title;
static string current_username;
static string previous_username;
// Timestamp of current entry; valid when we see contributor, comment, text
static string ts_current;
// Timestamp of previous entry
static string ts_previous;
/*
 * We need a temporary store for the revision's references,
 * because we can't asign them a time range until we see
 * another revision or the page's end.
 * Also we need to count each reference per period only once.
 */
class ReferenceDetails {
public:
	RangeMap range;			// When the reference has been used
	string firstRef;		// Timestamp of first referece
	bool needsUpdating;		// True if this entry's range needs updating
	ReferenceDetails(const string &ts) : range(), firstRef(ts), needsUpdating(true) {}
};
typedef map<string, ReferenceDetails> CurrentReferenceMap;
static CurrentReferenceMap current_references;
static set<string> current_contributors;
static int current_page_revisions;
static int current_page_reverts;

// Arguments for constructing a string from the contents of an (attributeless) tag
#define tagstring(s, tag) s + sizeof(tag) + 2 - 1, strlen(s) - (sizeof(tag) + 2 - 1) - (sizeof(tag) + 3 - 1)
// Explanation		       tag           <>  \0              tag           <>  \0    tag           </> \0

/*
 * Cannonicalize an entry's or reference's name
 * per http://en.wikipedia.org/wiki/Wikipedia:Canonicalization
 */
const string
canonicalize(const string &s)
{
	string result;
	string::const_iterator i;
	char c;
	bool seenspace = false;

	// strip leading spaces
	for (i = s.begin(); i != s.end() && (*i == ' ' || *i == '_'); i++)
		;

	// Capitalize first character
	if (i != s.end() && islower(*i)) {
		result += toupper(*i);
		i++;
	}

	for (; i != s.end(); i++) {
		switch (*i) {
		case '_':
		case ' ':
		case '\n':
			if (seenspace)
				break;
			seenspace = true;
			result += ' ';
			break;
		default:
			seenspace = false;
			result += *i;
			break;
		case '^':		// Strip illegal characters?
			break;
		}
	}
	// Strip trailing spaces
	string::reverse_iterator r;
	for (r = result.rbegin(); r != result.rend() && *r == ' '; r++)
		;
	result.erase(r.base(), result.end());
	return (result);
}

/*
 * Convert a timestamp into a time_t
 * Precondition: TZ is unset, so that we can use the portable mktime function
 * Times are in the format:
 * 2012-12-21T23:45:56
 */
time_t
ts_to_time(const string &ts)
{
	if (ts.empty())
		return (-1);
	struct tm t;
	istringstream timestamp(ts);
	char c;
	t.tm_sec = 0;		// Some timestamps don't supply seconds
	timestamp >> t.tm_year >> c >> t.tm_mon >> c >> t.tm_mday >> c >> t.tm_hour >> c >> t.tm_min >> c >> t.tm_sec;
	t.tm_mon--;
	t.tm_year -= 1900;
	t.tm_isdst = 0;
	return (mktime(&t));
}


/*
 * Change a string's colon characters with _
 * colon is our field separator
 */
const string
escape_colon(const string &s)
{
	string r(s);

	for (string::iterator i = r.begin(); i != r.end(); i++)
		if (*i == ':')
			*i = '_';
	return (r);
}

/*
 * Store the temporary values in the global store
 * The definitions are valid from ts_current to ts_end
 * Return an iterator pointing to the entry
 */
static EntryMap::iterator
record_results(const string &ts_end, const string &username)
{
	pair<EntryMap::iterator, bool> ir;
	// Deal with the definition
	const string &defname = canonicalize(current_title);
	ir = entries.insert(EntryMap::value_type(defname,
		EntryDetails(RangeMap(ts_current, ts_end), RangeMap(ts_current, ts_end, currently_stub))));
		//	     defined                       stub
	if (!ir.second) {
		// Not inserted; already exists in ir.first
		ir.first->second.stub.setActive(ts_current, ts_end, currently_stub);
		ir.first->second.defined.setActive(ts_current, ts_end);
	}
	if (ir.first->second.firstDef == -1) {
		ir.first->second.definer = username;
		ir.first->second.firstDef = ts_to_time(ts_current);
	}
	// Handle the references
	for (CurrentReferenceMap::iterator cref = current_references.begin(); cref != current_references.end(); cref++)
		if (cref->second.needsUpdating) {
			cref->second.range.setActive(ts_current, ts_end);
			cref->second.needsUpdating = false;
		}

	pair<ActivityMap::iterator, bool> ci;
	ci = contributors.insert(ActivityMap::value_type(username, RangeMap(ts_current)));
	if (!ci.second)
		ci.first->second.setActive(ts_current);
	return (ir.first);
}

extern "C" void
title(char *s)
{
	current_title = string(tagstring(s, "title"));
	cerr << current_title << endl;
}

extern "C" void
revert()
{
	if (current_page_reverts == 0)
		num_reverted++;
	current_page_reverts++;
	num_reverts++;
	if (!ts_previous.empty())
		revert_time += ts_to_time(ts_current) - ts_to_time(ts_previous);
	else
		cerr << "No previous timestamp for revert at " << ts_current << endl;
}

extern "C" void
timestamp(char *s)
{
	string ts(tagstring(s, "timestamp"));
	num_revisions++;
	// Is one set in place?
	if (!ts_current.empty())
		record_results(ts, previous_username);
	ts_previous = ts_current;
	ts_current = ts;
	currently_stub = false;
	currently_redirected = 0;
	current_page_revisions++;
}

extern "C" void
username(char *s)
{
	previous_username = current_username;
	current_username = string(tagstring(s, "username"));
	current_contributors.insert(current_username);
}

extern "C" void
move(char *s)
{
	string sOld, sNew;

	for (s += sizeof("<comment>moved [[") - 1; *s != ']'; s++)
		sOld += *s;
	for (s += sizeof("]] to [[") - 1; *s != ']'; s++)
		sNew += *s;
	moves.push_back(Move(sOld, sNew));
}

extern "C" void
reference(char *s)
{
	pair<CurrentReferenceMap::iterator, bool> ir =
		current_references.insert(CurrentReferenceMap::value_type(canonicalize(s), ReferenceDetails(ts_current)));
	if (!ir.second)
		// Not inserted; already exists in ir.first
		ir.first->second.needsUpdating = true;
}

extern "C" void
ip()
{
	previous_username = current_username;
	current_username.clear();
	num_ip++;
}

extern "C" void
redirect()
{
	currently_redirected = 1;
}

extern "C" void
stub()
{
	currently_stub = true;
}

extern "C" void
restrictions()
{
	num_restrictions++;
}

extern "C" void
page_end(void)
{
	EntryMap::iterator ir = record_results(RangeMap::END_OF_TIME, current_username);

	// Handle the references
	for (CurrentReferenceMap::const_iterator cref = current_references.begin(); cref != current_references.end(); cref++) {
		pair<EntryMap::iterator, bool> ir =
			entries.insert(EntryMap::value_type(cref->first, EntryDetails(RangeMap(), RangeMap())));
		//								defined     stub
		// For every locally set reference, increment the global entry's reference number
		for (int i = 0; i < RangeMap::NMONTH; i++)
			ir.first->second.nrefs[i] += cref->second.range.isActive(i);
		// Update first referrer details
		if (ir.first->second.firstRef.empty() || cref->second.firstRef < ir.first->second.firstRef) {
			ir.first->second.referrer = current_username;
			ir.first->second.firstRef = cref->second.firstRef;
		}
	}

	num_redirections += currently_redirected;
	num_pages++;

	// Update the entry's and the maximum numbers
	int nr = 0;
	for (CurrentReferenceMap::const_iterator i = current_references.begin(); i != current_references.end(); i++)
		nr += i->second.range.isActive(RangeMap::NMONTH - 1);
	ir->second.numReferences = nr;
	if (nr > max_references) {
		max_references = nr;
		max_references_title = current_title;
	}

	ir->second.numContributors = current_contributors.size();
	if (current_contributors.size() > max_contributors) {
		max_contributors = current_contributors.size();
		max_contributors_title = current_title;
	}

	ir->second.numRevisions = current_page_revisions;
	if (current_page_revisions > max_revisions) {
		max_revisions = current_page_revisions;
		max_revisions_title = current_title;
	}

	ir->second.numReverts = current_page_reverts;
	if (current_page_reverts > max_reverts) {
		max_reverts = current_page_reverts;
		max_reverts_title = current_title;
	}
}

extern "C" void
page_begin(void)
{
	current_references.clear();
	ts_current.clear();
	ts_previous.clear();
	current_contributors.clear();
	current_page_revisions = 0;
	current_page_reverts = 0;
}

extern "C" void yylex();

int
main(int argc, char *argv[])
{
	#ifdef unix
	setenv("TZ", "", 1);
	tzset();
	#endif
	yylex();
	cout << "% $Revision: 1.38 $" << endl;
	cout << "% Number of bins: " << RangeMap::NMONTH << endl;
	cout << "% Total revisions: " << num_revisions << endl;
	cout << "% Maximum revisions: " << max_revisions << " (" << max_revisions_title << ')' << endl;
	cout << "% Maximum reverts: " << max_reverts << " (" << max_reverts_title << ')' << endl;
	cout << "% Number of moves: " << moves.size() << endl;
	cout << "% Total pages: " << num_pages << endl;
	cout << "% Revisions from IP addresses: " << num_ip << endl;
	cout << "% Total contributors: " << contributors.size() << endl;
	cout << "% Maximum different contributors: " << max_contributors << " (" << max_contributors_title << ')' << endl;
	cout << "% Redirected pages: " << num_redirections << endl;
	cout << "% Restricted pages: " << num_restrictions << endl;
	cout << "% Maximum number of contained references: " << max_references << " (" << max_references_title << ')' << endl;
 	cout << "% Pages with at least one revert: " << num_reverted << endl;
	cout << "% Total number of reverts across all pages: " << num_reverts << endl;
	cout << "% Total time between reverts: " << revert_time << endl;

	/*
	 * When a page moves, its history goes to the new name
	 * and the history of its old name appears to start at the
	 * point of the move.
	 * Adjust the old pages to start at the same point as the
	 * new page.  This doesn't work across multiple moves,
	 * which loose the move history, and overestimates the old
	 * number of articles.
	 */
	int num_moves = 0;
	for (MoveContainer::const_iterator mi = moves.begin(); mi != moves.end(); mi++) {
		EntryMap::iterator from = entries.find(mi->sOld);
		if (from == entries.end())
			continue;
		EntryMap::iterator to = entries.find(mi->sNew);
		if (to == entries.end())
			continue;
		for (int i = 0; i < RangeMap::NMONTH; i++) {
			if (to->second.defined.isActive(i))
				from->second.defined.setActive(i);
			if (to->second.stub.isActive(i))
				from->second.stub.setActive(i);
		}
		num_moves++;
	}

	cout << "% Moved pages: " << num_moves << endl;
	// Calculate the set difference for every period
	// Print those of the last period
	EntryMap::iterator ei;
	ActivityMap::iterator contribi;
	vector<int> ndefs(RangeMap::NMONTH, 0);		// Number of definitions
	vector<int> nstubs(RangeMap::NMONTH, 0);	// Number of stub definitions
	vector<int> nrefa(RangeMap::NMONTH, 0);		// Number of referenced articles
	vector<int> nrefs(RangeMap::NMONTH, 0);		// Number of references
	vector<int> nurefs(RangeMap::NMONTH, 0);	// Number of undefined references
	vector<int> ncontribs(RangeMap::NMONTH, 0);	// Number of active contributors
	cout << "\n% name first-def first-definer first-ref first-referrer num-references num-contributors num-revisions num-reverts nrefs... E\n";
	for (ei = entries.begin(); ei != entries.end(); ei++) {
		cout << escape_colon(ei->first);
		cout << ':' << ei->second.firstDef << ':' << escape_colon(ei->second.definer);
		cout << ':' << ts_to_time(ei->second.firstRef) << ':' << escape_colon(ei->second.referrer);
		cout << ':' << ei->second.numReferences << ':' << ei->second.numContributors << ':' << ei->second.numRevisions;
		cout << ':' << ei->second.numReverts;
		// Count defs, refs, stubs each period
		for (int i = 0; i < RangeMap::NMONTH; i++) {
			if (ei->second.defined.isActive(i))
				ndefs[i]++;
			cout << ':' << ei->second.nrefs[i];
			if (ei->second.nrefs[i]) {
				nrefs[i] += ei->second.nrefs[i];
				nrefa[i]++;
				if (!ei->second.defined.isActive(i))
					nurefs[i]++;
			}
			if (ei->second.stub.isActive(i))
				nstubs[i]++;
		}
		cout << ":E" << endl;
	}

	// Calculate developers per period
	for (contribi = contributors.begin(); contribi != contributors.end(); contribi++)
		for (int i = 0; i < RangeMap::NMONTH; i++)
			if (contribi->second.isActive(i))
				ncontribs[i]++;


	cout << "\n% Period Entries Stubs References Ref-entries Undefined Contributors\n";
	for (int i = 0; i < RangeMap::NMONTH; i++)
		cout << RangeMap::periodName(i) << '\t' << ndefs[i] << '\t' << nstubs[i] << '\t' << nrefs[i] << '\t' << nrefa[i] << '\t' << nurefs[i] << '\t' << ncontribs[i] << endl;
	return (0);
}
